// Package
package org.auxilum.opengl.importing.milkshape;

// Standard
import java.io.*;
import java.util.StringTokenizer;

// Custom
import org.auxilum.opengl.scene.Model;
import org.auxilum.opengl.scene.Mesh;
import org.auxilum.opengl.scene.Material;


/**
 *
 *
 **/
public class MilkShapeModel implements Model
{
	/* Define field names for different elements in file */
	private final String MS3D_ASCII_MESHES = "Meshes:";
	private final String MS3D_ASCII_MATERIALS = "Materials:";
	

	// Model attributes
	private String name = null;
	
	private int meshCount = -1;
	private Mesh[] meshObjArray = null;
	
	private int materialCount = -1;
	private Material[] materialObjArray = null;
	
	private float[] position = null;
	private boolean doRender = true;

	/**
	 *	Constructor that takes a MilkShape ASCII file as it's argument
	 **/
	public MilkShapeModel(String milkshapeAsciiFile, String modelName)
	{
		this.name = modelName;
		
		try
		{
			parseFile(milkshapeAsciiFile);
		}
		catch (IOException exc)
		{
			new RuntimeException("There was an error reading " + milkshapeAsciiFile, exc);
		}
		
		//System.out.println("Model " + milkshapeAsciiFile + " loaded");
	}
	
	/**
	 *
	 **/
	public void setPosition(float[] position)
	{
		if (position.length != 3)
			throw new IllegalArgumentException("Position array must contain exactly 3 values (X,Y,Z)");
		this.position = position;
	}
	
	
	/**
	 *
	 **/
	public float[] getPosition()
	{
		return this.position;
	}


	/**
	 *
	 **/
	public void setDoRender(boolean doRender)
	{
		this.doRender = doRender;
	}
	
	
	/**
	 *
	 **/
	public boolean getDoRender()
	{
		return this.doRender;
	}	
	
	/**
	 *
	 **/
	public String getName()
	{
		return this.name;
	}

		
	/**
	 *	Returns the number of meshes in this model
	 **/
	public int getMeshCount()
	{
		return this.meshCount;
	}
	
	
	/**
	 *	Returns the meshes that this model consists of	
	 **/	
	public Mesh[] getMeshes()
	{
		return this.meshObjArray;
	}
	
	
	/**
	 *
	 **/
	public Material[] getMaterials()
	{
		return this.materialObjArray;
	}

	
	/**
	 *	This method read the Milkshape ASCII file and creates the <code>MilkShapeModel</code> object.
	 **/
	private void parseFile(String filename) throws FileNotFoundException, IOException
	{
		// Create file handle
		File file = new File(filename);
		
		// Open file
		LineNumberReader reader =  new LineNumberReader(new BufferedReader(new InputStreamReader(new FileInputStream(file))));
		String line = null;
		
		while(true)
		{
			// Read one line
			line = reader.readLine();
			
			if (line == null)
			{
				// We are done
				break;	
			}
			else if (line.length() > MS3D_ASCII_MESHES.length() && line.substring(0, MS3D_ASCII_MESHES.length()).equals(MS3D_ASCII_MESHES))
			{
				// Number of meshes
				this.meshCount = Integer.parseInt(line.substring(MS3D_ASCII_MESHES.length()).trim());
				
				//System.out.println("Meshes: " + meshCount);
				
				// Init mesh object array
				this.meshObjArray = new Mesh[meshCount];
				
				// Parse out all meshes
				int triangleCount = 0;
				for (int i = 0; i < meshCount; i++)
				{
					meshObjArray[i] = readMesh(reader);
					triangleCount += meshObjArray[i].getTriangleCoordIndexes().length;
				}
				
				//System.out.println("Model triangles: " + triangleCount); 
			}
			else if (line.length() > MS3D_ASCII_MATERIALS.length() && line.substring(0, MS3D_ASCII_MATERIALS.length()).equals(MS3D_ASCII_MATERIALS))
			{
				// Number of materials
				this.materialCount = Integer.parseInt(line.substring(MS3D_ASCII_MATERIALS.length()).trim());
				
				//System.out.println("Materials: " + materialCount);
				
				// Init material object array
				this.materialObjArray = new Material[materialCount];
				
				// Parse out all materials
				for (int i = 0; i < materialCount; i++)
				{
					materialObjArray[i] = readMaterial(reader);
				}
			}
		}
		
		// Close file
		reader.close();
		
		
		// Set correct material for each mesh
		for (int i = 0; i < meshObjArray.length; i++)
		{
			int index = meshObjArray[i].getMeshMaterialIndex();
			
			if (index > -1)
				meshObjArray[i].setMaterial(materialObjArray[index]);
			else
				meshObjArray[i].setMaterial(null);
		}
	}
	
	
	private Material readMaterial(LineNumberReader reader) throws IOException
	{
		String materialName = reader.readLine();
		materialName = materialName.substring(1, materialName.length() - 1);
		//System.out.println("Material name: " + materialName);
		
		StringTokenizer st = null;
		
		// Ambient
		st = new StringTokenizer(reader.readLine());
		float[] ambient = new float[]{Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken())};
		
		// Diffuse
		st = new StringTokenizer(reader.readLine());
		float[] diffuse = new float[]{Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken())};
		
		// Specular
		st = new StringTokenizer(reader.readLine());
		float[] specular = new float[]{Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken())};
		
		// Emissive
		st = new StringTokenizer(reader.readLine());
		float[] emissive = new float[]{Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken())};
		
		// Shininess
		float shininess = Float.parseFloat(reader.readLine());
		
		// Transparency
		float transparency = Float.parseFloat(reader.readLine());
		
		// Colormap
		String colorMap = normalizeImageName(reader.readLine());
		//System.out.println(colorMap);
		
		
		// Alphamap
		String alphaMap = reader.readLine();
		alphaMap = alphaMap.substring(1, alphaMap.length()-1);
		
		// Create material
		return new Material(materialName, ambient, diffuse, specular, emissive, shininess, transparency, colorMap /* e.g. texture */);
		
	}
	
	private String normalizeImageName(String path)
	{
		path = path.replace('\\', '/');
		int slashIndex = path.lastIndexOf("/");
		int dotIndex = path.lastIndexOf(".");
		
		if (slashIndex == -1)
			slashIndex = 1;
		else
			slashIndex += 1;
			
		if (dotIndex == -1)
			dotIndex = path.length() -1;
			
		return path.substring(slashIndex, dotIndex);
	}
	
	
	private Mesh readMesh(LineNumberReader reader) throws IOException
	{
		// Get mesh name, flag and material
		String line = reader.readLine();
		StringTokenizer st = new StringTokenizer(line);
		String meshName = st.nextToken();
		meshName = meshName.substring(1, meshName.length() - 1);
		int meshFlag = Integer.parseInt(st.nextToken());
		int meshMaterialIndex = Integer.parseInt(st.nextToken());
		//System.out.println("Mesh name: " + meshName);
		
		
		// Determine number of vertices
		int verticeCount = Integer.parseInt(reader.readLine());
		if (verticeCount < 0)
		{
			throw new RuntimeException("Invalid data for vertice count specified at line " + reader.getLineNumber());
		}
		
		
		// Prepare array to hold vertice coordinates 
		// (flags, x, y, z, u, v, bone index)
		double[][] coords = new double[verticeCount][7];
		
		// Start reading the actual vertices of this mesh
		int index = 0;
		while (index < verticeCount)
		{
			line = reader.readLine();
			if (line == null)
			{
				throw new RuntimeException("Unexpected end of file at line " + reader.getLineNumber());
			}

			// Parse line
			st = new StringTokenizer(line);
			int verticeFlag = Integer.parseInt(st.nextToken());
			double x = Double.parseDouble(st.nextToken());
			double y = Double.parseDouble(st.nextToken());
			double z = Double.parseDouble(st.nextToken());
			double u = Double.parseDouble(st.nextToken());
			double v = Double.parseDouble(st.nextToken());
			int boneIndex = Integer.parseInt(st.nextToken());
			
			coords[index][0] = verticeFlag;
			coords[index][1] = x;
			coords[index][2] = y;
			coords[index][3] = z;
			coords[index][4] = u;
			coords[index][5] = v;
			coords[index][6] = boneIndex;
			
			index++;
		}
		
		
		// Determine number of normals
		int normalsCount = Integer.parseInt(reader.readLine());
		
		// Prepare array to hold normal coordinates 
		// (x, y, z)
		double[][] normalCoords = new double[normalsCount][3];
		
		index = 0;
		while (index < normalsCount)
		{
			line = reader.readLine();
			if (line == null)
			{
				throw new RuntimeException("Unexpected end of file at line " + reader.getLineNumber());
			}

			// Parse line
			st = new StringTokenizer(line);
			double x = Double.parseDouble(st.nextToken());
			double y = Double.parseDouble(st.nextToken());
			double z = Double.parseDouble(st.nextToken());

			normalCoords[index][0] = x;
			normalCoords[index][1] = y;
			normalCoords[index][2] = z;
			index++;
		}
		
		
		// Determine number of triangles
		int triangleCount = Integer.parseInt(reader.readLine());
		
		// Prepare array to hold triangle coordinates 
		// (flags, vertex index1, vertex index2, vertex index3, normal index1, normal index 2, normal index 3, smoothing group)
		int[][] triangleCoords = new int[triangleCount][8];
		
		index = 0;
		while (index < triangleCount)
		{
			line = reader.readLine();
			if (line == null)
			{
				throw new RuntimeException("Unexpected end of file at line " + reader.getLineNumber());
			}

			// Parse line
			st = new StringTokenizer(line);
			int flag = Integer.parseInt(st.nextToken());
			int xIndex = Integer.parseInt(st.nextToken());
			int yIndex = Integer.parseInt(st.nextToken());
			int zIndex = Integer.parseInt(st.nextToken());

			triangleCoords[index][0] = flag;
			triangleCoords[index][1] = xIndex;
			triangleCoords[index][2] = yIndex;
			triangleCoords[index][3] = zIndex;
			index++;
		}
				
		
		
		// Create mesh object and return it
		return new Mesh(meshName, meshFlag, meshMaterialIndex, coords, normalCoords, triangleCoords);
	}
}
