package org.auxilum.opengl.robot;

// Standard
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;

import java.util.Map;
import java.util.TreeMap;


// Custom
import org.auxilum.opengl.common.SystemConfigFrame;


public class RobotFrame extends JFrame
{
	// Config
	private static int canvasWidth = 1024;
	private static int canvasHeight = 768;

	// Attributes
	private SceneRenderer renderer = null;
	private static RobotFrame frame = null;
	private SystemConfigFrame configFrame = null;
	private RobotControlPanel robotControlPanel = null;
	
	private Container container = null;

	private JMenuItem exitItem = null;
	
	private JButton btnSetup = null;
	private JLabel fpsLabel = null;
	private JCheckBox chkFpsContraint = null;
	private JSlider fpsSlider = null;
	
	private static final String windowTitle = " Robot Arm Simulator";

    private RobotStatus robotStatus = null;
    private DisplayMode[] displayModes = null;
    
    private GraphicsDevice device = null;
    private DisplayMode originalDM;
    private static boolean runningFullscreen = false;
                
	/** */
	public RobotFrame(GraphicsDevice device) throws Exception
	{
		super(device.getDefaultConfiguration());
		displayModes = device.getDisplayModes();
		for (int i = 0; i < displayModes.length; i++)
		{
			if 	(displayModes[i].getWidth() >= 640
				&& (displayModes[i].getHeight() * 100 / displayModes[i].getWidth()) == 75
				&& (displayModes[i].getBitDepth() == 16 || displayModes[i].getBitDepth() == 32)
				&& displayModes[i].getRefreshRate() >= 60 
				&& displayModes[i].getRefreshRate() <= 100 
				)
			{
	
				/*System.out.println("#" + i 
								+ ": " 
								+ displayModes[i].getWidth() 
								+ "x" 
								+ displayModes[i].getHeight() 
								+ "x" 
								+ displayModes[i].getBitDepth()
								+ " - " 
								+ displayModes[i].getRefreshRate());
				*/	
			}
		}
		
		// Save reference to rendering device		
		this.device = device;
		
		// Save current screen mode
		originalDM = device.getDisplayMode();

		// Setup window
		setupGUI();
    	
    	// Add KeyListener
      	KeyResponder keyResponder = new KeyResponder();
      	this.addKeyListener(keyResponder);
      	
      	// Init robot status
      	this.robotStatus = new RobotStatus(
      							new Axis[]{new Axis("Base", -166.0f, 166.0f, 0.0f), 
      							new Axis("Lower arm", -110.0f, 70.0f, 0.0f),
      							new Axis("Elbow", -64.0f, 190.0f, 0.0f),
      							new Axis("Upper arm", -180.0f, 180.0f, 0.0f)
      							});
      	
      	// Add control panel
    	JPanel robotPanel = new RobotControlPanel(robotStatus);
    	robotPanel.setMaximumSize(new Dimension(canvasWidth, 100));
    	container.add(robotPanel, BorderLayout.SOUTH);
    	
      	// Init renderer object
    	this.renderer = new SceneRenderer(robotStatus, robotControlPanel);
    	this.renderer.getCanvas().addKeyListener(keyResponder);
    	
    	// Add renderer to window
    	container.add(renderer.getCanvas(), BorderLayout.CENTER);
    	
		// Add windowslistener
    	this.addWindowListener(new WindowAdapter() 
    	{
        	public void windowClosing(WindowEvent e) 
        	{
          		exit();
        	}
      	});
	}
	
	private void switchMode(int modeNumber)
	{
		device.setFullScreenWindow(frame);
		//device.setDisplayMode(displayModes[modeNumber]);
		device.setDisplayMode(new DisplayMode(1024, 768, 32, 75));
	}
	
	/** Main method */
	public static void main(String[] args) throws Exception
	{
		System.out.println("Starting Robot arm simulator...");
		
		// Load custom look and feel
		System.out.print("Loading custom Look & Feel...");
		UIManager.setLookAndFeel(new com.incors.plaf.kunststoff.KunststoffLookAndFeel());
		System.out.println(" Done");

		// Check Jogl lib
		if (!Util.checkLib("jogl")) 
		{
			JOptionPane.showMessageDialog(null, "The JOGL OpenGL library was not found in 'java.library.path'", "Error", JOptionPane.ERROR_MESSAGE);
			System.exit(1);
		}
		
		// Check GageTimer lib
		if (!Util.checkLib("timer")) 
		{
			JOptionPane.showMessageDialog(null, "The Gage Timer (Native high res timer) library was not found in 'java.library.path'", "Error", JOptionPane.ERROR_MESSAGE);
			System.exit(1);
		}
		
		// Determine window/fullscreen
		int fullScreen = JOptionPane.showConfirmDialog(null, "Would you like to run in fullscreen mode?", "Fullscreen",  JOptionPane.YES_NO_CANCEL_OPTION);

		if (fullScreen == JOptionPane.CANCEL_OPTION)
		{
			System.exit(0);
		}


		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment(); 
	  	GraphicsDevice gd = ge.getDefaultScreenDevice(); 
		GraphicsConfiguration gc = gd.getDefaultConfiguration();
		
		// Create the window
		frame = new RobotFrame(gd); 
		
		boolean isFullScreen = gd.isFullScreenSupported();
        frame.setUndecorated(isFullScreen);
		frame.setSize(canvasWidth, canvasHeight); 

		if( isFullScreen && fullScreen == JOptionPane.YES_OPTION) 
		{
			runningFullscreen = true;
			frame.setResizable(false);
			
			// Switch resolution
			frame.switchMode(250);
   		}
   		else
   		{
   			frame.setUndecorated(false); 
   		}
   		
   		      	
      	// Show frame
    	frame.requestFocus();
    	frame.setVisible(true);
   		frame.show();
   		
  	}
  	
  	
  	
  	/** Prepare frame */
  	private void setupGUI()
  	{
  		this.container = new JPanel(new BorderLayout(0, 0));
  		//this.container.setBackground(Color.BLACK);
  		getContentPane().add(container);

  		setTitle(windowTitle);
    	setSize(canvasWidth, canvasHeight);
    	
    	setIconImage(new ImageIcon("img/icon.gif").getImage());
  	}
   	
   	
   	/** Used to exit program */
   	private void exit()
   	{
   		renderer.stop();
   		if (runningFullscreen)
   			device.setDisplayMode(originalDM);
        System.exit(0);
   	}
   	
   	

   	
   	
   	/***********************************************/
   	private class KeyResponder extends KeyAdapter
   	{
   		/** */
   		public void keyPressed(KeyEvent e)
   		{
   			int keyCode = e.getKeyCode();
   			switch(keyCode)
   			{
   				case KeyEvent.VK_ESCAPE:
   					exit();
   					break;
   					
   				case KeyEvent.VK_S:
   					if (configFrame == null)
   					{
   						// Create config window
    					configFrame = new SystemConfigFrame(renderer.getContextInfo(), renderer);
    					configFrame.setDefaultCloseOperation(HIDE_ON_CLOSE);
   					}
   					configFrame.show();
   					break;
   			}
   		}
   	}
}

