package org.auxilum.opengl.robot;

// Standard
import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;

// XPP xml pull parser
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;


public class RobotProgram 
{
	
	private int axisCount = -1;
	private int animationFps = 30;
	
	private String lastStart = null;
	
	int counter = 0;
	private int time = -1;
	private int[] IDs = null;
	private boolean[] absoluteMoves = null;
	
	private AxisMovement[] axisMovements = null;
	private ArrayList movesList = null;
	private double[][] moveData = null;
	
	/**
	 * 
	 **/
	public RobotProgram(String programPath) 
	{
		this.movesList = new ArrayList();
		
		// Read program file
		try
		{
			XmlPullParserFactory factory = XmlPullParserFactory.newInstance(System.getProperty(XmlPullParserFactory.PROPERTY_NAME), null);
			XmlPullParser xpp = factory.newPullParser();
			xpp.setInput(new BufferedReader(new FileReader(programPath)));
			processDocument(xpp);
		}
		catch (XmlPullParserException exc)
		{
			throw new RuntimeException(exc.getMessage(), exc);
		}
		catch (FileNotFoundException exc)
		{
			throw new RuntimeException(exc.getMessage(), exc);
		}
		catch (IOException exc)
		{
			throw new RuntimeException(exc.getMessage(), exc);
		}
		
		// Program structure has been parsed, "compile" it by building the multi-dimensional array
		this.moveData = compile(animationFps);
		
		//System.out.println("Length of program: " + moveData[0].length);
	}
	
	
	/**
	 *
	 **/
	public double[][] getMovements()
	{
		return this.moveData;
	}	
	
	
	private void processDocument(XmlPullParser xpp) throws XmlPullParserException, IOException
    {
        int eventType = xpp.getEventType();
        
        do {
            if(eventType == XmlPullParser.START_DOCUMENT) {
                //System.out.println("Start document");
            } else if(eventType == XmlPullParser.END_DOCUMENT) {
                //System.out.println("End document");
            } else if(eventType == XmlPullParser.START_TAG) {
                processStartElement(xpp);
            } else if(eventType == XmlPullParser.END_TAG) {
                processEndElement(xpp);
            } else if(eventType == XmlPullParser.TEXT) {
                processText(xpp);
            }
            eventType = xpp.next();
        } while (eventType != XmlPullParser.END_DOCUMENT);
    }
    
    
    private void processStartElement (XmlPullParser xpp)
    {
        String name = xpp.getName();
        
        if(name.equals("robotprogram"))
        {
        	this.axisCount = Integer.parseInt(xpp.getAttributeValue(0));
        }
        else if (name.equals("move"))
        {
        	time = Integer.parseInt(xpp.getAttributeValue(0));
        	
        	// Init arrays for data
        	IDs = new int[axisCount];
			absoluteMoves = new boolean[axisCount];
			axisMovements = new AxisMovement[axisCount];
        	
        }
        else if (name.equals("axis"))
        {
        	IDs[counter] = Integer.parseInt(xpp.getAttributeValue(0));
        	absoluteMoves[counter] = Integer.parseInt(xpp.getAttributeValue(1)) == 1;
        }
        
        
        // Save what is last start tag processed
        this.lastStart = name;
    }


    private void processEndElement (XmlPullParser xpp)
    {
        String name = xpp.getName();
        
        if (name.equals("move"))
        {
        	// A move section is done, create a Move object from the data parsed
        	Move move = new Move(time, axisMovements);
        	movesList.add(move);
        	counter = 0;
        }
        else if (name.equals("axis"))
        {
        	// An axis is parsed, increase counter
        	counter++;
        }
    }


    private void processText (XmlPullParser xpp) throws XmlPullParserException
    {
        String data = xpp.getText();
        
        if (data != null && lastStart.toLowerCase().equals("axis") && (!data.trim().equals("")) )
        {
        	String[] c = data.split(",");
        	double[] coords = new double[c.length];
        	for (int i = 0; i < c.length; i++)
        	{
	        	coords[i] = Double.parseDouble(c[i]);
	        }
	        
	        axisMovements[counter] = new AxisMovement(IDs[counter], absoluteMoves[counter], coords);
        }
    }
    
    
    /**
     *	Creates an array of positions used for doing the animation
     **/
    private double[][] compile(int aniFps)
    {
		// Determine total time for execution
		int totalTime = 0;
    	Iterator it = movesList.iterator();
    	Move move = null;
    	while (it.hasNext())
    	{
    		move = (Move) it.next();
    		totalTime += move.getTime();
    	}
    	
    	// Init array
		int initAnimationFrames = (int) (Math.ceil(((double)totalTime / 1000.0d) * aniFps) * 1.5);
		double[][] aniArr = new double[axisCount][initAnimationFrames];
    	
    	// Parse movements
    	it = movesList.iterator();
    	AxisMovement[] axisMovements = null;
    	int frameCounter = 0;
    	
    	while (it.hasNext())
    	{
    		move = (Move) it.next();
    		axisMovements = move.getAxisMovements();
    		int time = move.getTime();
    		
    		// Number of frames in this move
    		int frameCountInMove = (int) Math.ceil(((double)time / 1000d) * aniFps);
    		//System.out.println("Frame count in move: " + frameCountInMove);
    		
    		// Each axis
    		for (int axis = 0; axis < axisMovements.length; axis++)
    		{
    			AxisMovement axisMovement = axisMovements[axis];
    			
    			if (axisMovement != null)
    			{
	    			double[] positions = axisMovement.getPositions();
	    			double totalAngleMovement = 0.0;
	    			for (int i = 0; i < positions.length; i++)
	    			{
	    				totalAngleMovement += Math.abs(positions[i]);
	    			}
	    			
					// Determine movement speed of axis to get finished in time specified
    				double angleSpeed = totalAngleMovement / time;
    				
    				// Determine how much the arm must be moved per frame in degrees
    				double angleMovementPerFrame = angleSpeed / aniFps;
    				//System.out.println(angleMovementPerFrame);
    				
    				int framesUsed = 0;
    				for (int i = 0; i < positions.length; i++)
    				{
						// Frames for this position change
						int fpsForMove = (int) (time / (double)aniFps);
						double delta = Math.abs(positions[i]) / totalAngleMovement;
    					int framesForPositionChange = (int) Math.ceil(((double)time / 1000d) * aniFps * delta);
    					//System.out.println("Frames for position change: " + framesForPositionChange);
    					
  						// Set values in array for movement
						int a = frameCounter + framesUsed;
						int max = frameCounter + framesUsed + framesForPositionChange;
						for (; a < max; a++)
						{
							aniArr[axisMovement.getId()][a] = positions[i] / framesForPositionChange; //moveTo;
						}
						
						framesUsed+=framesForPositionChange;
    				}
    			}
    		}
    		
    		// Update total frames used
    		frameCounter += frameCountInMove;
    	}
    	
    	double[][] tmp = new double[axisCount][frameCounter];
      	for (int i = 0; i < axisCount; i++)
      	{
      		for (int j = 0; j < frameCounter; j++)
      		{
      			tmp[i][j] = aniArr[i][j];
      		}
      	}
    	aniArr = tmp;  	
    	//tmp = null;
	      	
    	// return array
    	return aniArr;
    }
    
    
    /**
     *
     **/
    private class Move
    {
		// Attributes
    	private int time = -1;
    	private AxisMovement[] axisMovements = null;
    	
    	
    	/**
    	 *
    	 **/
    	public Move(int time, AxisMovement[] axisMovements)
    	{
    		this.time = time;
    		this.axisMovements = axisMovements;
    	}
    	
    	/**
    	 *
    	 **/
    	public int getTime()
    	{
    		return this.time;
    	}
    	
    	/**
    	 *
    	 **/
    	public AxisMovement[] getAxisMovements()
    	{
    		return this.axisMovements;
    	}
    	
    }
    
    
    /**
     *
     **/
    private class AxisMovement
    {
    	// Attributes
    	private int axisId = -1;
    	private boolean abs = false;
    	private double[] moves = null;
    	
		/**
		 *
		 **/
    	public AxisMovement(int axisId, boolean abs, double[] moves)
    	{
    		this.axisId = axisId;
    		this.abs = abs;
    		this.moves = moves;
    	}
    	
    	/**
    	 *
    	 **/
    	public double[] getPositions()
    	{
    		return this.moves;
    	}
    	
    	/**
    	 *
    	 **/
    	public int getId()
    	{
    		return this.axisId;
    	}
    }
}
