package org.auxilum.opengl.robot;

// Standard
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

import java.util.Map;
import java.util.HashMap;

// JOGL
import net.java.games.jogl.*;
import net.java.games.jogl.util.GLUT;

// GAGE timer (Native high res timer)
import com.dnsalias.java.timer.*;

// Custom
import org.auxilum.opengl.scene.*;
import org.auxilum.opengl.importing.milkshape.*;


public class SceneRenderer implements GLEventListener, MouseListener, MouseMotionListener, MouseWheelListener
{
	// Used to keep track of robot movement
	private RobotStatus robotStatus = null;
	private RobotControlPanel robotControlPanel = null;
	
	// Config options
	private int targetFPS = 0;
	private boolean showFPS = true;
	
	// Context info
	private Map contextInfo = new HashMap();
	
	// Attributes used to keep track of free rotation by mouse
	private float zoom = -60.0f;
    private float	xrot;
    private float	yrot;
    private float	zrot;
    private int prevMouseX, prevMouseY;
	private boolean mouseRButtonDown = false;
	

    //*******************//
	
	private GL gl;
	private GLU glu;	
	private GLUT glut = null;
	private GLDrawable glDrawable;
	
	private int framesRendered = 0;
	private int fps = 0;
	
	private Animator animator = null;
	
	private AdvancedTimer timer = null;
	private long ticks = 0;
	
	// Canvas to draw the graphics on
    private GLCanvas canvas = null;
    
    // The main scene
    private Scene scene = null;
    
    
    /**
     *	Constructor 
     **/
   	public SceneRenderer(RobotStatus robotStatus, RobotControlPanel robotControlPanel)
   	{
   		// Save robot status reference
   		this.robotStatus = robotStatus;
   		this.robotControlPanel = robotControlPanel;
   		
		// Load models
		String modelPath = "data/models/";
		String[] modelFiles = new String[]
						{	
						"foot.txt", 
						"rotation.txt",
						"lower_arm.txt",
						"elbow.txt",
						"upper_arm.txt",
						"floor.txt",
						"mount.txt"
						};

		// Init scene
		this.scene = new Scene();
		
		
		System.out.print("Loading scene models");
   		for (int i = 0; i < modelFiles.length; i++)
   		{
	   		this.scene.addModel(new MilkShapeModel(modelPath + modelFiles[i], modelFiles[i]));
	   		System.out.print(".");
	   	}
	   	System.out.println(" Done");
	   	
		// Create OpenGL canvas
	   	canvas = GLDrawableFactory.getFactory().createGLCanvas(new GLCapabilities());
   		canvas.addGLEventListener(this);
   		
   		// Gage timer
   		this.timer = new AdvancedTimer();
   		this.timer.start();
   	}
    
    
	/**
 	 * 	Render scene 
 	 **/
	public void display(GLDrawable drawable) 
    {
		// Clear scene
      	gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
      	
      	// Reset view
      	gl.glLoadIdentity();

		// Update view if user has rotated or zoomed the scene    	
      	gl.glTranslatef(0.0f, -12.0f, zoom);
      	gl.glRotated(xrot, 1.0, 0.0, 0.0);
      	gl.glRotated(yrot, 0.0, 1.0, 0.0);     
      	
      	// Update light source
      	//gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, new float[]{-10f, -150f, 75 /*zoom * -1.5f*/, 1.0f}); 	

		/** Draw models **/

		// Draw robot mount
		gl.glPushMatrix();
		gl.glTranslatef(-7.2f, 0f, -4.6f);
		gl.glScalef(0.3f, 0.3f, 0.4f);
		drawModel(scene.getDisplayListIndex("mount.txt"), gl);
		
		// Foot
		gl.glPopMatrix();
		gl.glPushMatrix();
		//gl.glScalef(10f, 10f, 10f);
		drawModel(scene.getDisplayListIndex("foot.txt"), gl);
		
		
		// Rotation model
		gl.glTranslatef(0.0f, 2.0f, 0.0f);
		gl.glRotatef(robotStatus.getAxisAngle(0), 0.0f, 1.0f, 0.0f);
		drawModel(scene.getDisplayListIndex("rotation.txt"), gl);

					
		// Lower arm
		gl.glTranslatef(3.3f, 5.8f, 1.35f);
		gl.glRotatef(robotStatus.getAxisAngle(1), 0.0f, 0.0f, 1.0f);
		drawModel(scene.getDisplayListIndex("lower_arm.txt"), gl);
		
		// Elbow model
		gl.glTranslatef(0.0f, 12.85f, 1.0f);
		gl.glRotatef(robotStatus.getAxisAngle(2), 0.0f, 0.0f, 1.0f);
		drawModel(scene.getDisplayListIndex("elbow.txt"), gl);
		
		// Upper arm
		gl.glTranslatef(2.25f, 2.0f, -2.85f);
		gl.glRotatef(robotStatus.getAxisAngle(3), 1.0f, 0.0f, 0.0f);
		drawModel(scene.getDisplayListIndex("upper_arm.txt"), gl);

		// Draw the simple box
		gl.glPopMatrix();
		gl.glPushMatrix();
		//gl.glScalef(2.6f, 2.6f, 2.6f);
		gl.glTranslatef(10f, -3f, 10f);
		drawModel(scene.getDisplayListIndex("box.txt"), gl);
		
		// Draw industrial container
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glScalef(6f, 6f, 6f);
		gl.glTranslatef(8f, -1f, -60f);
		gl.glRotatef(-95f, 0f, 1f, 0f);
		drawModel(scene.getDisplayListIndex("container.txt"), gl);
		
		// Draw the floor
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glTranslatef(-150f, -5f, -350f);
		gl.glScalef(15f, 1f, 25f);
		drawModel(scene.getDisplayListIndex("floor.txt"), gl);
		
		
		// End
		gl.glPopMatrix();
		
		
		
		// Ensure scene is rendered completly
		gl.glFinish();


		// Used to keep track of FPS
		framesRendered++;
		
		
		// Try to run at target FPS
		if (targetFPS > 0)
		{
			long sleepTime = timer.getTicksPerSecond() / targetFPS;
    		timer.sleepUntil(ticks + sleepTime); 
    		ticks += sleepTime; 
    	}
    }
    
   	
   	/**
   	 *	Init scene
   	 **/
   	public void init (GLDrawable drawable) 
   	{
		// Save references
   		this.gl = drawable.getGL();
   		this.glu = drawable.getGLU();
      	this.glDrawable = drawable;
      	
   		// Prepare scene
   		this.scene.prepareScene(gl, glu);
   		
   		// Get context info
   		this.contextInfo.put("GL_VENDOR", gl.glGetString(GL.GL_VENDOR));
   		this.contextInfo.put("GL_RENDERER", gl.glGetString(GL.GL_RENDERER));
   		this.contextInfo.put("GL_VERSION", gl.glGetString(GL.GL_VERSION));

		// OpenGL rendering options		
		gl.glShadeModel(GL.GL_SMOOTH);              // Enable Smooth Shading
		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.5f);    // Black Background
		gl.glClearDepth(1.0f);                      // Depth Buffer Setup
		gl.glEnable(GL.GL_DEPTH_TEST);							// Enables Depth Testing
		gl.glDepthFunc(GL.GL_LEQUAL);								// The Type Of Depth Testing To Do
		gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);	// Really Nice Perspective Calculations
		
		// Texture filtering options
		gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR); 
		gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR); 


		// Enable lightning
		float[] light_position = {0.0f, 150.0f, -20.0f, 1.0f};
		float[] white_light = {1.0f, 1.0f, 1.0f, 1.0f};
		float[] amb_light = {0.2f, 0.2f, 0.2f, 1.0f};
		
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, light_position);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, white_light);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, amb_light);

		gl.glEnable(GL.GL_LIGHT0);
		gl.glEnable(GL.GL_LIGHTING);

		// Add mouse listener        
		drawable.addMouseListener(this);
		drawable.addMouseMotionListener(this);
		drawable.addMouseWheelListener(this);
		
		   		
		// Start rendering updater thread
		animator = new Animator(getCanvas());
		animator.start();
		
		// Start FPS timer
		FPSTimer fpsTimer = new FPSTimer();
		new Thread(fpsTimer).start();
	}

   
    public void reshape(GLDrawable drawable, int x, int y, int width, int height) 
    {
		gl.glLoadIdentity();

		// Avoid a divide by zero
		if (height <= 0) 
        	height = 1;
        
        // GL code
		final float h = (float)width / (float)height;
		gl.glViewport(0, 0, width, height);
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();
		glu.gluPerspective(45.0f, h, 1.0, 1000.0);
		gl.glMatrixMode(GL.GL_MODELVIEW);
	}


	
    
    /** */
    public void displayChanged(GLDrawable drawable, boolean modeChanged, boolean deviceChanged) 
    {
    	
    }

	
	public void mouseWheelMoved(MouseWheelEvent e)
	{
		zoom += (e.getWheelRotation() / 2.0f);
	} 

	
	
    // Methods required for the implementation of MouseListener
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}

    public void mousePressed(MouseEvent e) 
    {
		prevMouseX = e.getX();
      	prevMouseY = e.getY();
      	if ((e.getModifiers() & e.BUTTON3_MASK) != 0) 
      	{
        	mouseRButtonDown = true;
      	}
    }
    
    
    public void mouseReleased(MouseEvent e) 
    {
    	if ((e.getModifiers() & e.BUTTON3_MASK) != 0) 
    	{
        	mouseRButtonDown = false;
      	}
    }
    
    public void mouseClicked(MouseEvent e) 
    {
    }
    
    
    // Methods required for the implementation of MouseMotionListener
    public void mouseDragged(MouseEvent e) 
    {
    	int x = e.getX();
      	int y = e.getY();
      	Dimension size = e.getComponent().getSize();

      	float thetaY = 360.0f * ( (float)(x-prevMouseX)/(float)size.width);
      	float thetaX = 360.0f * ( (float)(prevMouseY-y)/(float)size.height);
    
      	prevMouseX = x;
      	prevMouseY = y;

      	xrot += thetaX;
      	yrot += thetaY;
    }
    
    
    public void mouseMoved(MouseEvent e) 
    {
    }
   	
   	
   	/**
   	 *
   	 **/
   	public void setTargetFPS(int targetFPS)
   	{
   		this.targetFPS = targetFPS;
   	}

	
	/**
	 *
	 **/
	public int getMaxFPS()
	{
		return targetFPS;
	}

	
	/**
	 *
	 **/
	public int getFPS()
	{
		return fps;
	}
	
	
	/**
	 *
	 **/
	public GLCanvas getCanvas()
	{
		return canvas;
	}
	
	
	/**
	 *
	 **/
	public Map getContextInfo()
	{
		return this.contextInfo;
	}

	
	/**
	 *
	 **/	
	public void stop()
	{
		if (animator != null)
			animator.stop();
	}
    
    
    /**
     *
     **/
    private void drawModel(int index, GL gl)
    {
    	gl.glCallList(index+1);
    }
    
    private class FPSTimer implements Runnable
	{
		// Delay for updating the FPS display
		private int delay = 500;

		public void run()
		{
			while (true)
			{
				fps = (int) (1000.0d * framesRendered / delay);
				framesRendered = 0;				
				
				// Sleep before updating
				try
				{
					Thread.sleep(delay);
				}			
				catch (InterruptedException exc)
				{
					// Ignore
				}
			}
		}
	}
}