// Package
package org.auxilum.opengl.scene;

// Standard
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

// JOGL
import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Custom
import org.auxilum.opengl.scene.util.TextureList;


public class Scene
{
	
	// Attributes
	private Map models = null;
	private TextureList textureList = null;
	private boolean sceneReady = false;
	private Map displayListIndex = null;

	
	/**
	 *	Default constructor
	 **/
	public Scene()
	{
		this.models = new HashMap();
		this.displayListIndex = new HashMap();
	}
	
	
	/**
	 *	Adds a new model to the scene
	 **/
	public void addModel(Model model)
	{
		if (models.containsKey(model.getName()))
		{
			throw new IllegalArgumentException("Model " + model.getName() + " already exists in scene");
		}
		this.models.put(model.getName(), model);
	}
	
	
	/**
	 *	Returns the model specified by modelName
	 **/
	public Model getModel(String modelName)
	{
		return (Model) models.get(modelName);
	}
	
	/**
	 *
	 **/
	public boolean unloadModel(String modelName)
	{
		return models.remove(modelName) != null;
	}
	
	
	/**
	 *
	 **/
	public void prepareScene(GL gl, GLU glu)
	{
		// Load textures
		loadTextures(gl, glu);

		// Create display lists
		createDispLists(gl);
		
		// The scene can now be rendered
		this.sceneReady = true;
	}
	
	
	/**
     *
     **/
    private void createDispLists(GL gl)
    {
    	System.out.print("Creating display lists");
    	// Init displaylists
    	int dispList = gl.glGenLists(models.size());
    	
    	// Get iterator for model list
    	Iterator iter = models.values().iterator();
    	int index = 0;
    	while (iter.hasNext())
    	{
    		Model curModel = (Model) iter.next();
			Mesh[] meshes = curModel.getMeshes();
		
			System.out.print(".");
		
			// Init new list
			gl.glNewList(index + 1, GL.GL_COMPILE);				
			
			int m;
			for (m = 0; m < meshes.length; m++)
			{
				// Draw each mesh
				int[][] triangleIndexes = meshes[m].getTriangleCoordIndexes();
				double[][] vertices = meshes[m].getVerticeCoords();
				double[][] normals = meshes[m].getNormals();
				Material material = meshes[m].getMaterial();
				
				
				// Specify texture
				if (material != null)
				{
					//System.out.println("Material: " + material);
					Texture texture = material.getTexture();
					
					if (texture != null)
					{
						gl.glEnable(GL.GL_TEXTURE_2D);
						gl.glBindTexture(GL.GL_TEXTURE_2D, texture.getGLTextureId());
					}
					else
					{
						//System.out.println("Disabling textures");
						gl.glDisable(GL.GL_TEXTURE_2D);
					}
					
					// Specify material properties
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT, material.getAmbient());
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_DIFFUSE, material.getDiffuse());
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, material.getSpecular());
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_EMISSION, material.getEmission());
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, new float[]{material.getShininess()});
				}
				else
				{
					float[] defaultArr = new float[]{0.2f, 0.2f, 0.2f};
					
					gl.glDisable(GL.GL_TEXTURE_2D);
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT, defaultArr);
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_DIFFUSE, defaultArr);
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, defaultArr);
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_EMISSION, defaultArr);
					gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, new float[]{0.0f});
				}
				
				
				
				for (int i = 0; i < triangleIndexes.length; i++)
				{
					// Draw each polygon
					gl.glBegin(GL.GL_TRIANGLES);
		
						// Get vertice indexes
						int xIndex = triangleIndexes[i][1];
						int yIndex = triangleIndexes[i][2];
						int zIndex = triangleIndexes[i][3];
						
						int xNormalIndex = triangleIndexes[i][4];
						int yNormalIndex = triangleIndexes[i][5];
						int zNormalIndex = triangleIndexes[i][6];
						
						// Specify normals
						gl.glNormal3d(normals[xNormalIndex][0], normals[yNormalIndex][1], normals[zNormalIndex][2]);
						
						// Specify texture coords and triangles
						//gl.glTexCoord2d(0, 0); //    vertices[xIndex][1], vertices[xIndex][2] /*, vertices[xIndex][3]*/);
						gl.glVertex3d(vertices[xIndex][1], vertices[xIndex][2], vertices[xIndex][3]);
						
						//gl.glTexCoord2d(2,2); //vertices[yIndex][1], vertices[yIndex][2] /*, vertices[yIndex][3]*/);
						gl.glVertex3d(vertices[yIndex][1], vertices[yIndex][2], vertices[yIndex][3]);
						
						//gl.glTexCoord2d(0, 0); //(vertices[zIndex][1], vertices[zIndex][2], vertices[zIndex][3]);
						gl.glVertex3d(vertices[zIndex][1], vertices[zIndex][2], vertices[zIndex][3]);
						
					gl.glEnd();
				}
			}
			
			gl.glEndList();
			
			displayListIndex.put(curModel.getName(), new Integer(index));
			
			index++;
		}
		
		System.out.println(" Done");
	}
	
	
	/**
	 *
	 **/
	private TextureList loadTextures(GL gl, GLU glu)
	{
		TextureList textureList = new TextureList();
		Iterator iter = models.values().iterator();
		Model curModel = null;
		while (iter.hasNext())
		{
			curModel = (Model) iter.next();
			Material[] materials = curModel.getMaterials();
			
			if (materials != null)
			{
				for (int i = 0; i < materials.length; i++)
				{
					String textureName = materials[i].getTextureName();
					
					if (textureName != null && !textureName.equals(""))
					{
						System.out.println("Loading texture: " + textureName);
						textureList.addTexture(new Texture(textureName, gl, glu));
						materials[i].setTexture(textureList.getTexture(textureName));
					}
						
				}
			}
		}
		
		return textureList;
	}
	
	/**
	 *
	 **/
	public int getDisplayListIndex(String model)
	{
		Integer i = (Integer) this.displayListIndex.get(model);
		if (i != null)
			return i.intValue();
		else
			return -1;
	}
}