// Package
package org.auxilum.opengl.scene;

/** 
 *	This class represents a texture.
 *	It tries to hide all the gory details needed to initialize things so they work with JOGL.
 *  Based upon code from Kevin Duling's implementation of NeHe lesson 6.
 **/

import java.awt.*;
import java.io.*;
import java.util.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import java.nio.*;
import javax.imageio.*;

import net.java.games.jogl.*;
import net.java.games.jogl.util.*;

public class Texture
{
    // Attributes
    int[] texName = new int[1];
    String filename = null;

    public Texture(String filename, GL gl, GLU glu) 
    {
    	if (filename == null || filename.equals(""))
    	{
    		return;
    	}
	    
	    gl.glTexParameteri( GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT );
	    gl.glTexParameteri( GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT );
	    gl.glTexParameteri( GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR );
	    gl.glTexParameteri( GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR );
	    
   		this.filename = filename;
	    String pathname = "data/textures/" + filename + ".png";
	    gl.glGenTextures( 1, texName );
	    gl.glBindTexture( GL.GL_TEXTURE_2D, texName[0] );
	    BufferedImage img = readPNGImage(pathname);
	    makeRGBTexture(gl, glu, img, GL.GL_TEXTURE_2D, false);

	    
    }
  
  	
	private BufferedImage readPNGImage(String resourceName)
	{
		try
		{
			URL url = getResource(resourceName);
			if (url == null)
			{
			  	throw new RuntimeException("Error reading resource " + resourceName);
			}
			BufferedImage img = ImageIO.read(url);
			int h = img.getHeight(null);
			int w = img.getWidth(null);
			//scale to next power of 2
			int minsize = Math.max(h,w);
			double size;

			for (size = 32; size < minsize; size *= 2); 
			
			java.awt.geom.AffineTransform tx = java.awt.geom.AffineTransform.getScaleInstance(size/w, -size/h);
			tx.translate(0, -h); 
			AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_NEAREST_NEIGHBOR); 
			img = op.filter(img, null); 
			return img;
		}
		catch (IOException e)
		{
			throw new RuntimeException("Error reading resource " + resourceName + "\nDetails: " + e);
		}
	}
    
    private void makeRGBTexture(GL gl, GLU glu, BufferedImage img, int target, boolean mipmapped)
    {
      ByteBuffer dest = null;
      switch (img.getType())
      {
        case BufferedImage.TYPE_3BYTE_BGR:
        case BufferedImage.TYPE_CUSTOM:
        {
          byte[] data = ((DataBufferByte) img.getRaster().getDataBuffer()).getData();
          dest = ByteBuffer.allocateDirect(data.length);
          dest.order(ByteOrder.nativeOrder());
          dest.put(data, 0, data.length);
          break;
        }
        case BufferedImage.TYPE_INT_RGB:
        {
          int[] data = ((DataBufferInt) img.getRaster().getDataBuffer()).getData();
          dest = ByteBuffer.allocateDirect(data.length * BufferUtils.SIZEOF_INT);
          dest.order(ByteOrder.nativeOrder());
          dest.asIntBuffer().put(data, 0, data.length);
          break;
        }
        default:
          throw new RuntimeException("Unsupported image type " + img.getType());
      }
      
      if (mipmapped)
      {
        glu.gluBuild2DMipmaps(target, GL.GL_RGB8, img.getWidth(), img.getHeight(), GL.GL_RGB, GL.GL_UNSIGNED_BYTE, dest);
      }
      else
      {
        gl.glTexImage2D(target, 0, GL.GL_RGB, img.getWidth(), img.getHeight(), 0, GL.GL_RGB, GL.GL_UNSIGNED_BYTE, dest);
      }
    }


	/** 
	 *	Retrieve a URL resource from the jar.  If the resource is not found, then
	 * 	the local disk is also checked.
	 * 	@param filename Complete filename, including parent path
	 * 	@return a URL object if resource is found, otherwise null.
	 */  
	private URL getResource(final String filename)
	{
		// Try to load resource from jar
		URL url = ClassLoader.getSystemResource(filename);
		// If not found in jar, then load from disk
		if (url == null)
		{
			try
			{
				url = new URL("file", "localhost", filename);
			}
			catch (Exception urlException)
			{
				// Ignore	
			} 
		}
		return url;
	}


    /**
     * 	Name (integer id) of the texture, suitable for putting in <code>glBindTexture</code>
     **/
    public int getGLTextureId() 
    {
		return texName[0];
    }
    
	
	/**
	 *
	 **/
	public String getTextureName()
	{
		return this.filename;
	}
	    
	/** 
	 *	The texture file name 
	 **/
    public String toString()
    {
		return "Texture:\nFilename=" +  this.filename + "\nGL name (id)=" + texName[0];
    }

}
